import json
from collections import Counter
from typing import Dict, Optional

import pandas
import tqdm
from nltk import RegexpParser, word_tokenize

from wordview import logger
from wordview.mwes.am import calculate_am
from wordview.mwes.mwe_utils import get_pos_tags, is_alphanumeric_latinscript_multigram


class MWE(object):
    """
    Represents a Multiword Expression.
    """

    def __init__(
        self,
        df: pandas.DataFrame,
        text_column: str,
        mwe_types: list[str] = ["NC"],
        tokenize=False,
    ) -> None:
        """Initialize a new MWE object with the given df, text_column and mwe_types.

        Args:
            df (pandas.DataFram): DataFrame with a text_column that contains the corpus.
            text_column (str): Specifies the column of DataFrame where text data resides.
            mwe_types (List): Types of MWEs to be extracted. Supports: NC for Noun-Noun and JNC for Adjective-Noun compounds. Example: ['NC', 'JNC'].
            tokenize (bool): Tokenize the content of `df[text_column]`.

        Returns:
            None
        """
        self.df = df
        self.text_column = text_column
        if not mwe_types:
            raise ValueError("mwe_types is empty.")
        if not isinstance(mwe_types, list):
            raise TypeError("mwe_types is not a list.")
        for mt in mwe_types:
            if mt not in ["NC", "JNC"]:
                raise ValueError(f"{mt} type is not recognized.")
        self.mwe_types = mwe_types
        if tokenize:
            logger.info(
                '"tokenize" flag set to True. This might lead to a slow instantiation.'
            )
            self.df[text_column] = self.df[text_column].apply(self._tokenize)
        else:
            self._check_tokenized()

    def _tokenize(self, x):
        """Helper function to tokenize and join the results with a space.

        Args:
            x:

        Returns:
            None
        """
        return " ".join(word_tokenize(x))

    def _check_tokenized(self) -> None:
        """Helper function to check if the content of text_column is tokenized.

        Args:
            None

        Returns:
            None
        """
        if self.df[self.text_column].shape[0] > 200:
            tests = self.df[self.text_column].sample(n=200).tolist()
        else:
            tests = self.df[self.text_column].sample(frac=0.8).tolist()
        num_pass = 0
        for t in tests:
            try:
                if " ".join(word_tokenize(t)) == t:
                    num_pass += 1
            except Exception as E:
                logger.error(f"Could not tokenize and join tokens in {t}: \n {E} ")

        if float(num_pass) / float(len(tests)) < 0.8:
            logger.warning(
                f"It seems that the content of {self.text_column} in the input data frame is not (fully) tokenized.\nThis can lead to poor results. Consider re-instantiating your MWE instance with 'tokenize' flag set to True.\nNote that this might lead to a slower instantiation."
            )

    def build_counts(self, counts_filename: Optional[str] = None) -> Optional[Dict]:
        """Create various count files to be used by downstream methods
        by calling wordview.mwes.mwe_utils.

        Args:
            counts_filename (str): Filename for storing counts.

        Returns:
            None when no counts_filename is provided, otherwise res which is a dictionary of counts.
        """
        logger.info("Creating counts...")
        res = self.get_counts()
        if not counts_filename:
            return res
        else:
            try:
                with open(counts_filename, "w") as file:
                    json.dump(res, file)
            except Exception as e:
                logger.error(e)
                raise e
            return None

    def extract_mwes(
        self,
        am: str = "pmi",
        mwes_filename: Optional[str] = None,
        counts_filename: Optional[str] = None,
        counts: Optional[dict] = None,
    ) -> dict:
        """
        Extract MWEs from counts_filename with respect to the association measure specified by `am`.

        Args:
            am (str): The association measure to be used. Can be any of [pmi, npmi]
            mwes_filename (str): File for storing MWEs. Defaults to None.
            counts_filename (str): File to read counts from.

        Returns:
            Dictionary of MWEs.
        """
        if counts:
            count_data = counts
        elif counts_filename is not None:
            try:
                with open(counts_filename, "r") as file:
                    count_data = json.load(file)
            except Exception as e:
                logger.error(e)
                logger.error(
                    "Counts must be provided either via input argument `counts` or `counts_filename`. Argument `counts` is not specified and it seems like there was an error reading the counts from `counts_filename`."
                )
                raise e
        else:
            raise ValueError("Either 'counts' or 'counts_filename' must be provided.")

        logger.info(f"Extracting {self.mwe_types} based on {am}")
        mwe_am_dict = calculate_am(
            count_data=count_data, am=am, mwe_types=self.mwe_types
        )
        if mwes_filename:
            try:
                with open(mwes_filename, "w") as file:
                    json.dump(mwe_am_dict, file)
            except Exception as e:
                logger.error(e)
                raise e
            finally:
                return mwe_am_dict
        else:
            return mwe_am_dict

    def get_counts(self) -> Dict:
        """Read a corpus in pandas.DataFrame format and generates all counts necessary for calculating AMs.

        Args:
            None

        Returns:
            res: Dictionary of mwe_types to dictionary of individual mwe within that type and their count.
                E.g. {'NC':{'climate change': 10, 'brain drain': 3}, 'JNC': {'black sheep': 3, 'red flag': 2}}
        """
        res: Dict = {}
        for mt in self.mwe_types:
            res[mt] = {}
        res["WORDS"] = {}
        for sent in tqdm.tqdm(self.df[self.text_column]):
            tokens = sent.split(" ")
            word_count_dict = Counter(tokens)
            for k, v in word_count_dict.items():
                if k in res["WORDS"]:
                    res["WORDS"][k] += v
                else:
                    res["WORDS"][k] = v
            for mt in self.mwe_types:
                mwes_count_dic = self.extract_mwes_from_sent(tokens, mwe_type=mt)
                for k, v in mwes_count_dic.items():
                    if k in res[mt]:
                        res[mt][k] += v
                    else:
                        res[mt][k] = v
        return res

    def extract_mwes_from_sent(self, tokens: list[str], mwe_type: str) -> Dict:
        """Extract two-word noun compounds from tokenized input.

        Args:
            tokens: A tokenized sentence, i.e. list of tokens.
            type: Type of MWE. Any of ['NC', 'JNC'].

        Returns:
            mwes_count_dic: Dictionary of compounds to their count.
        """
        if not isinstance(tokens, list):
            raise TypeError(
                f'Input argument "tokens" must be a list of string. Currently it is of type {type(tokens)} \
                with a value of: {tokens}.'
            )
        if len(tokens) == 0:
            return {}
        mwes = []
        postag_tokens: list[tuple[str, str]] = get_pos_tags(tokens)
        w1_pos_tags = []
        w2_pos_tags = []
        if mwe_type == "NC":
            w1_pos_tags = ["NN", "NNS"]
            w2_pos_tags = ["NN", "NNS"]
        elif mwe_type == "JNC":
            w1_pos_tags = ["JJ"]
            w2_pos_tags = ["NN", "NNS"]
        for i in range(len(postag_tokens) - 1):
            w1 = postag_tokens[i]
            if w1[1] not in w1_pos_tags:
                continue
            else:
                w2 = postag_tokens[i + 1]
                if not is_alphanumeric_latinscript_multigram(
                    w1[0]
                ) or not is_alphanumeric_latinscript_multigram(w2[0]):
                    continue
                if w2[1] in w2_pos_tags:
                    if i + 2 < len(postag_tokens):
                        w3 = postag_tokens[i + 2]
                        if w3 not in ["NN", "NNS"]:
                            mwes.append(w1[0] + " " + w2[0])
                    else:
                        mwes.append(w1[0] + " " + w2[0])
        mwes_count_dic = Counter(mwes)
        return mwes_count_dic


class HigherOrderMWEExtractor:
    def __init__(self, tokens: list[str], pattern: str) -> None:
        self.tokens = tokens
        self.pattern = pattern
        self._validate_input()

    def _validate_input(self) -> None:
        if not isinstance(self.tokens, list):
            raise TypeError(
                f'Input argument "tokens" must be a list of string. Currently it is of type {type(self.tokens)} \
                with a value of: {self.tokens}.'
            )
        if len(self.tokens) == 0:
            raise ValueError(
                'Input argument "tokens" must be a non-empty list of string.'
            )
        if not isinstance(self.pattern, str):
            raise TypeError(
                f'Input argument "pattern" must be a string. Currently it is of type {type(self.pattern)} \
                with a value of: {self.pattern}.'
            )
        if len(self.pattern) == 0:
            raise ValueError(
                'Input argument "pattern" must be a non-zero length string.'
            )

    def extract_higher_order_mwe_candidates(self) -> dict:
        """
        Extract variable-length MWE from tokenized input, using a user-defined POS regex pattern.

        Parameters:
            tokens (list[str]): A list of tuples containing the word and its corresponding part-of-speech tag.
            pattern (str): A string containing a user-defined pattern for nltk.RegexpParser.

        Returns:
            match_counter (dict[str, dict[str, int]]): A counter dictionary with count of matched strings, grouped by pattern label.
                                                    An empty list if none were found.

        Examples of user-defined patterns:
        - NP: {<DT>?<JJ>*<NN>} # Noun phrase
        - VP: {<MD>?<VB.*><NP|PP|CLAUSE>+$} # Verb phrase
        - PP: {<IN><NP>} # Prepositional phrase

        You can use multiple and/or nested patterns, separated by a newline character:
        pattern = '''
        NP: {<DT>?<JJ>*<NN>} # Noun phrase
        PROPN: {<NNP>+} # Proper noun
        ADJP: {<RB|RBR|RBS>*<JJ>} # Adjective phrase
        ADVP: {<RB.*>+<VB.*><RB.*>*} # Adverb phrase
        '''

        In this case, patterns of a clause are executed in order.  An earlier
        pattern may introduce a chunk boundary that prevents a later pattern from executing.
        """

        tagged_tokens: list[tuple[str, str]] = get_pos_tags(self.tokens)
        parser = RegexpParser(self.pattern)
        parsed_tokens = parser.parse(tagged_tokens)

        labels: list[str] = [
            rule.split(":")[0].strip() for rule in self.pattern.split("\n") if rule
        ]

        matches: dict[str, list[str]] = {label: [] for label in labels}

        for subtree in parsed_tokens.subtrees():
            label = subtree.label()
            if label in matches:
                matches[label].append(
                    " ".join(word for (word, tag) in subtree.leaves())
                )

        matches_counter: dict[str, dict[str, int]] = {
            label: dict(Counter(match_list)) for label, match_list in matches.items()
        }
        return matches_counter



import json
import re
import sys
from collections import Counter
from re import Match
from typing import Any, Dict, List, Optional

import nltk
import pandas
import tqdm

from wordview import logger


def hyphenate_mwes(
    path_to_mwes: str,
    mwe_types: List[str],
    df: pandas.DataFrame,
    text_column: str,
    am_threshold: float = 0.7,
    only_mwes: bool = False,
    lower_case: bool = False,
) -> pandas.DataFrame:
    """Hyphenates the mwes in the corpus so that they are treated as a single token by downstream applications.

    Args:
        path_to_mwes: Path to a json file that contains a dictionary of MWE type for each type,
                    unique MWEs to their count. E.g. {'NC': {'mwe1': 10}}.
        mwe_types: Types of MWEs to be replaced. Can be any of [NC, JNC].
        df: DataFrame comprising training data with a tokenized text column.
        text_column: Text (content) column of df.
        am_threshold: MWEs with an am greater than or equal to this threshold are selected for replacement.
        only_mwes: Whether or not keep only MWEs and drop the rest of the text.
        lower_case: Whether or not lowercase the sentence before replacing MWEs.

    Returns:
        df (pandas.FataFrame)
    """
    try:
        with open(path_to_mwes, "r") as file:
            mwe_type_mwe_am = json.load(file)
    except Exception as e:
        raise e
    good_mwes = set()
    for t in mwe_types:
        logger.info(f"Replacing MWEs of type {t} in the corpus.")
        if t not in mwe_type_mwe_am:
            logger.error(
                f"MWEs of type {t} do not exist. Make sure the file containing MWEs in {path_to_mwes} includes type {t}"
            )
            sys.exit(1)
        pmi_sorted_dict = mwe_type_mwe_am[t]
        logger.info(f"Number of all MWEs of type {t}: {len(pmi_sorted_dict)}")
        for k, v in pmi_sorted_dict.items():
            if v >= am_threshold:
                good_mwes.add(k)
            else:
                break
        logger.info(
            "Number of MWEs to be replaced in corpus based on the association threshold: %d"
            % len(good_mwes)
        )

    logger.info("Replacing compounds in text")
    new_text = []
    for sent in tqdm.tqdm(df[text_column]):
        sent = sent.lower() if lower_case else sent
        bigrams = get_ngrams(sent, 2)
        if only_mwes:
            tmp = ""
            for bg in bigrams:
                if bg in good_mwes:
                    tmp += bg.split(" ")[0] + "-" + bg.split(" ")[1] + " "
            if tmp != "":
                sent = tmp.strip()
        else:
            for bg in bigrams:
                if bg in good_mwes:
                    sent = re.sub(bg, bg.split(" ")[0] + "-" + bg.split(" ")[1], sent)
        new_text.append(sent)
    df[text_column] = new_text
    return df


def get_ngrams(sentence: str, n: int) -> List[str]:
    """Extracts n-grams from sentence.

    Args:
        sentence: Input sentence from which n-grams are to be extracted.
        n: Size of n-grams.

    Returns:
        ngrams: List of extracted n-grams.
    """
    if not isinstance(sentence, str):
        raise TypeError(
            f"Input argument 'sentence' must be of type str. You have provided an input of type: {type(sentence)}."
        )
    ngrams: List[str] = []
    try:
        tokens = sentence.split(" ")
    except Exception as E:
        logger.error(E)
        logger.error(
            f'Input "{sentence}" cannot be spilitted around space. No n-gram is extracted.'
        )
        return ngrams
    for i in range(len(tokens) - n + 1):
        ngrams.append(" ".join(tokens[i : i + n]))
    return ngrams


def get_pos_tags(tokens: list[str]) -> list[tuple[str, str]]:
    pos_tags = nltk.pos_tag(tokens)
    return pos_tags


def is_alphanumeric_latinscript_multigram(word: str) -> Optional[Match[str]]:
    match: Optional[Match] = re.match("[a-zA-Z0-9]{2,}", word)
    return match



from typing import Dict, List, Set, Tuple

import pandas
import plotly.figure_factory as ff
import plotly.graph_objs as go
from tabulate import tabulate  # type: ignore

from wordview.text_analysis.core import (
    do_txt_analysis,
    generate_label_plots,
    plotly_wordcloud,
)


class TextStatsPlots:
    """
    Represents Text Statistics and Plots.
    """

    def __init__(
        self,
        df: pandas.DataFrame,
        text_column: str,
        distributions: Set = {"doc_len", "word_frequency_zipf"},
        pos_tags: Set = {"NN", "VB", "JJ"},
    ) -> None:
        """Initialize a new TextStatsPlots object with the given arguments.

        Args:
            df (pandas.DataFrame): DataFrame with a text_column that contains the corpus.
            text_column (str): Specifies the column of DataFrame where text data resides.
            distributions (Set): Set of distribution types to generate and plot. Available distributions are:
                doc_len: document lengths, word_frequency_zipf: Zipfian word frequency distribution.
                Default = {"doc_len", "word_frequency_zipf"}.
            pos_tags (Set): = Set of POS tags for which world cloud is shown. Default = {"NN", "VB", "JJ"}.

        Returns:
            None
        """
        self.df = df
        self.analysis = do_txt_analysis(df=self.df, text_col=text_column)
        self.distributions = distributions
        self.pos_tags = pos_tags
        self.languages = self.analysis.languages
        self.type_count = self.analysis.type_count
        self.token_count = self.analysis.token_count
        self.num_docs = self.analysis.doc_count
        self.median_doc_len = self.analysis.median_doc_len
        self.num_nns = len(self.analysis.nns)
        self.num_jjs = len(self.analysis.jjs)
        self.num_vbs = len(self.analysis.vs)

    def create_dist_plots(self) -> Dict[str, go.Figure]:
        """Create distribution plots for items in `self.distributions`.

        Args:
            None

        Returns:
            Dictionary of distribution names to plotly go.Figure objects for that distribution.
        """
        res = {}
        if "doc_len" in self.distributions:
            fig_doc_len_dist = ff.create_distplot(
                [self.analysis.doc_lengths], group_labels=["distplot"], colors=["blue"]
            )
            res["doc_len"] = fig_doc_len_dist

        if "word_frequency_zipf" in self.distributions:
            fig_w_freq = go.Figure()
            # Alternative nice color scales that go together:
            # Plotly3
            # ice
            fig_w_freq.add_trace(
                go.Scattergl(
                    x=self.analysis.zipf_x,
                    y=self.analysis.zipf_y_emp,
                    mode="markers",
                    marker=dict(
                        color=self.analysis.zipf_x,
                        colorscale="Tealgrn",
                    ),
                )
            )
            fig_w_freq.add_trace(
                go.Scattergl(
                    x=self.analysis.zipf_x,
                    y=self.analysis.zipf_y_theory,
                    mode="markers",
                    marker=dict(color=self.analysis.zipf_x, colorscale="Reds"),
                )
            )
            res["word_frequency_zipf"] = fig_w_freq

        dist_plot_setup = {
            # 'paper_bgcolor': '#007A78',
            "showlegend": False
        }
        for _, fig in res.items():
            fig.update_layout(dist_plot_setup)

        return res

    def show_distplot(self, distribution: str) -> None:
        """Shows distribution plots for `dist`.

        Args:
            dist (str): The distribution for which the plot is to be shown.
                        Can be either of: doc_len" or "word_frequency_zipf.

        Returns:
            None
        """
        self.create_dist_plots()[distribution].show()

    def create_pos_plots(
        self, go_plot_settings: Dict = {}, wc_settings: Dict = {}
    ) -> Dict[str, go.Figure]:
        """Create plots for the POS tags specified in items in `self.pos_tags`.

        Args:
            go_plot_settings (Dict): Color and other settings for the plotly.graph_objs figure.
            wc_settings (Dict): Color, font and other settings for wordcloud.WordCloud.

        Returns:
            Dictionary of POS tags to plotly go.Figure objects.

        """
        word_cloud_mandatory_settings = {
            "showlegend": False,
            "xaxis_showgrid": False,
            "yaxis_showgrid": False,
            "xaxis_zeroline": False,
            "yaxis_zeroline": False,
            "yaxis_visible": False,
            "yaxis_showticklabels": False,
            "xaxis_visible": False,
            "xaxis_showticklabels": False,
        }
        word_cloud_setting = {**word_cloud_mandatory_settings, **go_plot_settings}
        res = {}
        if "NN" in self.pos_tags:
            res["noun_cloud"] = go.Figure(
                plotly_wordcloud(
                    token_count_dic=self.analysis.nns, settings=wc_settings
                )
            )
        if "JJ" in self.pos_tags:
            res["adj_cloud"] = go.Figure(
                plotly_wordcloud(
                    token_count_dic=self.analysis.jjs, settings=wc_settings
                )
            )
        if "VB" in self.pos_tags:
            res["verb_cloud"] = go.Figure(
                plotly_wordcloud(token_count_dic=self.analysis.vs, settings=wc_settings)
            )
        for _, fig in res.items():
            fig.update_layout(word_cloud_setting)
        return res

    def show_word_clouds(
        self, pos: str, go_plot_settings: Dict = {}, wc_settings: Dict = {}
    ) -> None:
        """Shows POS word clouds.

        Args:
            pos (str): Type of POS. Can be any of: [NN, JJ, VB].
            go_plot_settings (Dict): Color and other settings for the word clouds.
            wc_settings (Dict): Color, font and other settings for wordcloud.WordCloud.

        Returns:
            None
        """
        if pos == "NN":
            self.create_pos_plots(
                go_plot_settings=go_plot_settings, wc_settings=wc_settings
            )["noun_cloud"].show()
        if pos == "JJ":
            self.create_pos_plots(
                go_plot_settings=go_plot_settings,
            )["adj_cloud"].show()
        if pos == "VB":
            self.create_pos_plots(
                go_plot_settings=go_plot_settings,
            )["verb_cloud"].show()

    def show_stats(self) -> None:
        """Print dataset statistics, including:
        Language/s
        Number of unique words
        Number of all words
        Number of documents
        Median document length
        Number of nouns
        Number of adjectives
        Number of verbs.
        """
        table = tabulate(
            [
                ["Language/s", ", ".join(self.languages)],
                ["Unique Words", f"{self.type_count:,d}"],
                ["All Words", f"{self.token_count:,d}"],
                ["Documents", f"{self.num_docs:,d}"],
                ["Median Doc Length", self.median_doc_len],
                ["Nouns", f"{self.num_nns:,d}"],
                ["Adjectives", f"{self.num_jjs:,d}"],
                ["Verbs", f"{self.num_vbs:,d}"],
            ],
            tablefmt="simple_grid",
        )
        print(table)

    def show_insights(self):
        "Show topics, MWEs, clusters,"
        raise NotImplementedError


class LabelStatsPlots:
    """
    Represents Label Statistics and Plots.
    """

    def __init__(
        self,
        df: pandas.DataFrame,
        label_columns: List[Tuple],
    ) -> None:
        """Initialize a new LabelStatsPlots object with the given arguments.

        Args:
            df (pandas.DataFrame): DataFrame with one or more label column/s.
            label_columns (List): List of tuples (column_name, label_type) that specify a label column and its type (categorical or numerical).

        Returns:
            None
        """
        self.df = df
        self.labels_fig = generate_label_plots(self.df, label_cols=label_columns)

    def show_label_plots(self) -> None:
        """Renders label plots for columns specified in `self.label_columns`.

        Args:
            None

        Returns:
            None
        """
        self.labels_fig.show()
